# -*- coding=utf-8 -*-

"""Heaviest presence score path PuLP ILP."""

from typing import Optional

from pulp import LpMaximize, LpProblem, lpSum

from khloraascaf.ilp.pulp_circuit import (
    circuit_from_the_starter_forward,
    flow_definition,
    intermediate_in_circuit,
)
from khloraascaf.ilp.pulp_repeated_fragments import (
    fix_repeats_subpaths,
    occurrences_priority,
)
from khloraascaf.ilp.pulp_var_db import PuLPVarPresScoreModel
from khloraascaf.multiplied_doubled_contig_graph import (
    CIND_IND,
    PRESSCORE_ATTR,
    MDCGraph,
    OccOrCT,
)
from khloraascaf.result import ScaffoldingResult


# ============================================================================ #
#                                  PULP MODEL                                  #
# ============================================================================ #
def best_presscore_model(mdcg: MDCGraph, starter_vertex: OccOrCT,
                         fix_result: Optional[ScaffoldingResult] = None) -> (
        tuple[LpProblem, PuLPVarPresScoreModel]):
    """Best path with highest presence score PuLP model.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    fix_result : ScaffoldingResult, optional
        Previous scaffolding result, by default `None`

    Returns
    -------
    LpProblem
        ILP problem
    PuLPVarPresScoreModel
        ILP variables
    """
    # ------------------------------------------------------------------------ #
    # Constants
    # ------------------------------------------------------------------------ #
    big_m = mdcg.multiplied_card() // 2

    # ------------------------------------------------------------------------ #
    # Problem
    # ------------------------------------------------------------------------ #
    prob = LpProblem(name='best_presscore', sense=LpMaximize)

    # ------------------------------------------------------------------------ #
    # Variables
    # ------------------------------------------------------------------------ #
    var = PuLPVarPresScoreModel(mdcg, starter_vertex)

    # ------------------------------------------------------------------------ #
    # Objective function
    # ------------------------------------------------------------------------ #
    __presscore_objective(prob, var, mdcg)

    # ------------------------------------------------------------------------ #
    # Constraints
    # ------------------------------------------------------------------------ #
    #
    # Path constraints
    #
    flow_definition(prob, var, big_m, mdcg)
    circuit_from_the_starter_forward(prob, var, mdcg, starter_vertex)
    intermediate_in_circuit(prob, var, mdcg, starter_vertex)
    #
    # Fix repeats sub-paths
    #
    if fix_result is not None:
        fix_repeats_subpaths(prob, var, fix_result)
    #
    # Speed-up
    #
    occurrences_priority(prob, var, mdcg)
    return prob, var


# ============================================================================ #
#                              OBJECTIVE FUNCTION                              #
# ============================================================================ #
def __presscore_objective(prob: LpProblem, var: PuLPVarPresScoreModel,
                          mdcg: MDCGraph):
    """Invf model objective function.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarPresScoreModel
        Variable data object
    mdcg : MDCGraph
        Multiplied doubled contig graph
    """
    vertices = mdcg.vertices()
    prob += lpSum(
        vertices.attr(v[CIND_IND], PRESSCORE_ATTR) * var.i[v]
        for v in var.i
    )



# -*- coding=utf-8 -*-

"""Solve integer linear programs module."""

from __future__ import annotations, division

from collections.abc import Iterable
from pathlib import Path
from typing import Optional

from pulp import GUROBI_CMD, PULP_CBC_CMD, LpProblem

from khloraascaf.exceptions import WrongSolverName
from khloraascaf.inputs import SOLVER_CBC, SOLVER_GUROBI


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                   Log File                                   #
# ---------------------------------------------------------------------------- #
SOLVER_LOG_PREFIX = 'solver'
"""Prefix of the solver log file name."""

LOG_EXT = 'log'
"""Extension of the solver log file."""


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                 Solve Models                                 #
# ---------------------------------------------------------------------------- #
def solve_pulp_problem(prob: LpProblem, solver: str,
                       log_path: Optional[Path] = None):
    """Instanciate and solve the PuLP model.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    solver : str
        MILP solver to use ('cbc' or 'gurobi')
    log_path : Path, optional
        Solver log file path, by default `None`

    Raises
    ------
    WrongSolverName
        The solver name is not correct
    """
    if solver == SOLVER_CBC:
        SolverCmd = PULP_CBC_CMD
    elif solver == SOLVER_GUROBI:
        SolverCmd = GUROBI_CMD
    else:
        raise WrongSolverName(solver)
    prob.solve(SolverCmd(msg=0, logPath=log_path))


# ---------------------------------------------------------------------------- #
#                                Logs Formatters                               #
# ---------------------------------------------------------------------------- #
def fmt_solver_log_name(solver: str,
                        instance_name: str,
                        ilp_combination: Iterable[str]) -> str:
    """Format solver log file name.

    Parameters
    ----------
    solver : str
        Solver name
    instance_name : str
        Instance name
    ilp_combination : iterable of str
        ILP string code combination

    Returns
    -------
    str
        Formatted filename
    """
    return (
        f'{SOLVER_LOG_PREFIX}_{solver}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{LOG_EXT}'
    )



# -*- coding=utf-8 -*-

"""Command Line Interface library and configs."""

from argparse import ArgumentParser, Namespace

from khloraascaf.inputs import (
    INSTANCE_NAME_DEF,
    MULT_UPB_DEF,
    OUTDIR_DEF,
    PRESSCORE_UPB_DEF,
    SOLVER_CBC,
    SOLVER_GUROBI,
    MultT,
    PresScoreT,
)


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                   Arguments                                  #
# ---------------------------------------------------------------------------- #
ARG_CONTIG_ATTRS = 'contig_attributes'
_ARG_CONTIG_ATTRS_TYPE = str  # pylint: disable=invalid-name
_ARG_CONTIG_ATTRS_HELP = (
    'The multiplicities and the presence score of the contigs'
)

ARG_CONTIG_LINKS = 'contig_links'
_ARG_CONTIG_LINKS_TYPE = str  # pylint: disable=invalid-name
_ARG_CONTIG_LINKS_HELP = 'The links between the oriented contigs'

ARG_CONTIG_STARTER = 'starter_id'
_ARG_CONTIG_STARTER_TYPE = str  # pylint: disable=invalid-name
_ARG_CONTIG_STARTER_HELP = 'The identifier of the starter contig'

# ---------------------------------------------------------------------------- #
#                                    Options                                   #
# ---------------------------------------------------------------------------- #
OPT_MULT_UPB = 'multiplicity_upperbound'
_OPT_MULT_UPB_FLAG = ('--mult-upbound',)
_OPT_MULT_UPB_TYPE = MultT  # pylint: disable=invalid-name
_OPT_MULT_UPB_HELP = 'The upper bound for unknown multiplicities'

OPT_PRESSCORE_UPB = 'presence_score_upperbound'
_OPT_PRESSCORE_UPB_FLAG = ('--presscore-upbound',)
_OPT_PRESSCORE_UPB_TYPE = PresScoreT  # pylint: disable=invalid-name
_OPT_PRESSCORE_UPB_HELP = 'The upper bound for unknown presence score'

OPT_SOLVER = 'solver'
_OPT_SOLVER_FLAG = ('--solver',)
_OPT_SOLVER_TYPE = str  # pylint: disable=invalid-name
_OPT_SOLVER_CHOICES = (SOLVER_CBC, SOLVER_GUROBI)
_OPT_SOLVER_HELP = 'The MILP solver to use'

OPT_OUTDIR = 'output_directory'
_OPT_OUTDIR_FLAG = ('--out-directory',)
_OPT_OUTDIR_TYPE = str  # pylint: disable=invalid-name
_OPT_OUTDIR_HELP = 'The output directory path, generated if not given'

OPT_INSTANCE_NAME = 'instance_name'
_OPT_INSTANCE_NAME_FLAG = ('--instance-name',)
_OPT_INSTANCE_NAME_TYPE = str  # pylint: disable=invalid-name
_OPT_INSTANCE_NAME_HELP = 'Custom prefix for outputs'

OPT_OUTDEBUG = 'debug'
_OPT_OUTDEBUG_FLAG = ('--debug',)
_OPT_OUTDEBUG_HELP = 'Outputs debugs'


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                              Command Line Parser                             #
# ---------------------------------------------------------------------------- #
def arguments_cmdline() -> Namespace:
    """Set arguments in command line.

    Returns
    -------
    Namespace
        Parsed command line
    """
    # ------------------------------------------------------------------------ #
    # Programm
    # ------------------------------------------------------------------------ #
    argparser = ArgumentParser(
        prog='python3.9 -m khloraascaf',
        description='Scaffolding python module of chloroplast genome assembly',
    )

    # ------------------------------------------------------------------------ #
    # Arguments
    # ------------------------------------------------------------------------ #
    argparser.add_argument(
        ARG_CONTIG_ATTRS,
        type=_ARG_CONTIG_ATTRS_TYPE,
        help=_ARG_CONTIG_ATTRS_HELP,
    )
    argparser.add_argument(
        ARG_CONTIG_LINKS,
        type=_ARG_CONTIG_LINKS_TYPE,
        help=_ARG_CONTIG_LINKS_HELP,
    )
    argparser.add_argument(
        ARG_CONTIG_STARTER,
        type=_ARG_CONTIG_STARTER_TYPE,
        help=_ARG_CONTIG_STARTER_HELP,
    )

    # ------------------------------------------------------------------------ #
    # Options
    # ------------------------------------------------------------------------ #
    argparser.add_argument(
        *_OPT_MULT_UPB_FLAG,
        dest=OPT_MULT_UPB,
        type=_OPT_MULT_UPB_TYPE,
        default=MULT_UPB_DEF,
        help=_OPT_MULT_UPB_HELP,
    )
    argparser.add_argument(
        *_OPT_PRESSCORE_UPB_FLAG,
        dest=OPT_PRESSCORE_UPB,
        type=_OPT_PRESSCORE_UPB_TYPE,
        default=PRESSCORE_UPB_DEF,
        help=_OPT_PRESSCORE_UPB_HELP,
    )
    argparser.add_argument(
        *_OPT_SOLVER_FLAG,
        dest=OPT_SOLVER,
        type=_OPT_SOLVER_TYPE,
        choices=_OPT_SOLVER_CHOICES,
        default=SOLVER_CBC,
        help=_OPT_SOLVER_HELP,
    )
    argparser.add_argument(
        *_OPT_OUTDIR_FLAG,
        dest=OPT_OUTDIR,
        type=_OPT_OUTDIR_TYPE,
        default=OUTDIR_DEF,
        help=_OPT_OUTDIR_HELP,
    )
    argparser.add_argument(
        *_OPT_INSTANCE_NAME_FLAG,
        dest=OPT_INSTANCE_NAME,
        type=_OPT_INSTANCE_NAME_TYPE,
        default=INSTANCE_NAME_DEF,
        help=_OPT_INSTANCE_NAME_HELP,
    )
    argparser.add_argument(
        *_OPT_OUTDEBUG_FLAG,
        dest=OPT_OUTDEBUG,
        action='store_true',
        help=_OPT_OUTDEBUG_HELP,
    )
    return argparser.parse_args()



# -*- coding=utf-8 -*-

"""The PuLPVarDB classes file."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator, Union

from pulp import LpBinary, LpContinuous, LpVariable
from pulp.pulp import LpAffineExpression
from revsymg.index_lib import IndexT

from khloraascaf.ilp.dirf_sets import (
    DirFT,
    adirf,
    adirf_other,
    consecutive_pdirf,
    dirf,
    dirf_builder,
    forbidden_nested,
    pdirf,
    pdirf_alpha,
)
from khloraascaf.ilp.invf_sets import (
    InvFT,
    ainvf,
    ainvf_other,
    consecutive_pinvf,
    forbidden_intersection,
    invf,
    invf_builder,
    pinvf,
    pinvf_alpha,
)
from khloraascaf.multiplied_doubled_contig_graph import (
    CIND_IND,
    COCC_IND,
    EOccOrCT,
    MDCGraph,
    OccOrCT,
)


# DOCU sphinx doc pulp var db
# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
BIN_THRESHOLD = 0.5


# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                Path Variables                                #
# ---------------------------------------------------------------------------- #
DiVarT = dict[OccOrCT, LpVariable]
DxVarT = dict[EOccOrCT, LpVariable]
DfVarT = dict[EOccOrCT, LpVariable]

# ---------------------------------------------------------------------------- #
#                                InvF Variables                                #
# ---------------------------------------------------------------------------- #
DmInvFVarT = dict[InvFT, LpVariable]
DintersVarT = dict[tuple[IndexT, InvFT, InvFT], LpVariable]

# ---------------------------------------------------------------------------- #
#                                DirF Variables                                #
# ---------------------------------------------------------------------------- #
DmDirFVarT = dict[DirFT, LpVariable]
DnestedVarT = dict[tuple[int, DirFT, DirFT], LpVariable]

# ---------------------------------------------------------------------------- #
#                                Alpha Variables                               #
# ---------------------------------------------------------------------------- #
DalphaVarT = dict[EOccOrCT, LpVariable]

# ---------------------------------------------------------------------------- #
#                              Adjacent Fragments                              #
# ---------------------------------------------------------------------------- #
DisadjVarT = dict[EOccOrCT, LpVariable]


# ============================================================================ #
#                        PULP VARIABLE DATA BASE CLASSES                       #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                 Base Classes                                 #
# ---------------------------------------------------------------------------- #
@dataclass
class PuLPVarPath():
    """Data class for necessary variables to modelise a path."""

    def __init__(self, mdcg: MDCGraph, starter_vertex: OccOrCT):
        """The Initializer."""
        self.i: DiVarT = {
            v: LpVariable(f'i_{v}', cat=LpContinuous)
            for v in mdcg.multiplied_vertices()
            if v[CIND_IND] != starter_vertex[CIND_IND]
        }
        self.x: DxVarT = {
            e: LpVariable(f'x_{e}', cat=LpBinary)
            for e in mdcg.multiplied_edges()
        }
        self.f: DfVarT = {
            e: LpVariable(f'f_{e}', lowBound=0, cat=LpContinuous)
            for e in mdcg.multiplied_edges()
        }


@dataclass
class PuLPVarInvF():
    """Data class for necessary variables to modelise inverted fragments."""

    repeat_frag_fn = invf
    repeat_frag_builder_fn = invf_builder
    pair_repeat_alpha_fn = pinvf_alpha
    pair_repeat_frag_fn = pinvf
    pair_repeat_consecutive_fn = consecutive_pinvf
    adj_repeat_frag_fn = ainvf
    adj_repeat_other_fn = ainvf_other
    forbidden_pairing_fn = forbidden_intersection

    def __init__(self, mdcg: MDCGraph):
        """The Initializer."""
        self.m: DmInvFVarT = {
            p: LpVariable(f'm_{p}', cat=LpBinary)
            for p in PuLPVarInvF.repeat_frag_fn(mdcg)
        }
        self._alpha: DalphaVarT = {
            (u, v): LpVariable(f'alpha_{u}_{v}', cat=LpBinary)
            for u, v in PuLPVarInvF.pair_repeat_alpha_fn(mdcg)
        }
        self.pairing_ban: DintersVarT = {
            (k, p, q): LpVariable(f'forbid_{k}_{p}_{q}', cat=LpBinary)
            for p, q in PuLPVarInvF.pair_repeat_frag_fn(mdcg)
            for k in range(8)
        }
        self.isadj: DisadjVarT = {
            (u, v): LpVariable(f'isadj_{u}_{v}', cat=LpBinary)
            for u, v in PuLPVarInvF.adj_repeat_frag_fn(mdcg)
        }

    def alpha(self, u: OccOrCT, v: OccOrCT) -> LpAffineExpression:
        r"""Return alpha variable according to u and v.

        Parameters
        ----------
        u : OccOrCT
            Multiplied oriented contig u
        v : OccOrCT
            Multiplied oriented contig v

        Returns
        -------
        LpAffineExpression
            binary variable alpha

        Notes
        -----
        Define that :math:`\alpha_{uv} = 1 - \alpha_{vu}`
        using only one :math:`(u, v)` pair key
        """
        if alpha_key(u, v):
            return self._alpha[u, v]
        return 1 - self._alpha[v, u]

    def invf_solution(self) -> Iterator[InvFT]:
        """Iterate over the inverted fragments solution.

        Yields
        ------
        InvFT
            Inverted fragments solution
        """
        for p, var_p in self.m.items():
            if var_p.value() > BIN_THRESHOLD:
                yield p


@dataclass
class PuLPVarDirF():
    """Data class for necessary variables to modelise direct fragments."""

    repeat_frag_fn = dirf
    repeat_frag_builder_fn = dirf_builder
    pair_repeat_alpha_fn = pdirf_alpha
    pair_repeat_frag_fn = pdirf
    pair_repeat_consecutive_fn = consecutive_pdirf
    adj_repeat_frag_fn = adirf
    adj_repeat_other_fn = adirf_other
    forbidden_pairing_fn = forbidden_nested

    def __init__(self, mdcg: MDCGraph):
        """The Initializer."""
        self.m: DmDirFVarT = {
            p: LpVariable(f'm_{p}', cat=LpBinary)
            for p in PuLPVarDirF.repeat_frag_fn(mdcg)
        }
        self._alpha: DalphaVarT = {
            (u, v): LpVariable(f'alpha_{u}_{v}', cat=LpBinary)
            for u, v in PuLPVarDirF.pair_repeat_alpha_fn(mdcg)
        }
        self.pairing_ban: DnestedVarT = {
            (k, p, q): LpVariable(f'forbid_{k}_{p}_{q}', cat=LpBinary)
            for k in range(8)
            for p, q in PuLPVarDirF.pair_repeat_frag_fn(mdcg)
        }
        self.isadj: DisadjVarT = {
            (u, v): LpVariable(f'isadj_{u}_{v}', cat=LpBinary)
            for u, v in PuLPVarDirF.adj_repeat_frag_fn(mdcg)
        }

    def alpha(self, u: OccOrCT, v: OccOrCT) -> LpAffineExpression:
        r"""Return alpha variable according to u and v.

        Parameters
        ----------
        u : OccOrCT
            Multiplied oriented contig u
        v : OccOrCT
            Multiplied oriented contig v

        Returns
        -------
        LpAffineExpression
            binary variable alpha

        Notes
        -----
        Define that :math:`\alpha_{uv} = 1 - \alpha_{vu}`
        using only one :math:`(u, v)` pair key
        """
        if alpha_key(u, v):
            return self._alpha[u, v]
        return 1 - self._alpha[v, u]

    def dirf_solution(self) -> Iterator[DirFT]:
        """Iterate over the direct fragments solution.

        Yields
        ------
        DirFT
            Direct fragments solution
        """
        for p, var_p in self.m.items():
            if var_p.value() > BIN_THRESHOLD:
                yield p


# ---------------------------------------------------------------------------- #
#                                 Model Classes                                #
# ---------------------------------------------------------------------------- #
@dataclass
class PuLPVarInvFModel(PuLPVarPath, PuLPVarInvF):
    """Data class for necessary variables to modelise invf model."""

    def __init__(self, mdcg: MDCGraph, starter_vertex: OccOrCT):
        PuLPVarPath.__init__(self, mdcg, starter_vertex)
        PuLPVarInvF.__init__(self, mdcg)


@dataclass
class PuLPVarDirFModel(PuLPVarPath, PuLPVarDirF):
    """Data class for necessary variables to modelise dirf model."""

    def __init__(self, mdcg: MDCGraph, starter_vertex: OccOrCT):
        PuLPVarPath.__init__(self, mdcg, starter_vertex)
        PuLPVarDirF.__init__(self, mdcg)


@dataclass
class PuLPVarPresScoreModel(PuLPVarPath):
    """Data class for necessary variables to modelise presscore model."""

    def __init__(self, mdcg: MDCGraph, starter_vertex: OccOrCT):
        PuLPVarPath.__init__(self, mdcg, starter_vertex)


PuLPVarRepFModelT = Union[PuLPVarInvFModel, PuLPVarDirFModel]
PuLPVarModelT = Union[PuLPVarRepFModelT, PuLPVarPresScoreModel]


# ============================================================================ #
#                                ALPHA VARIABLES                               #
# ============================================================================ #
def alpha_key(u: OccOrCT, v: OccOrCT) -> bool:
    """Return the oriented unredondant (u, v) alpha key.

    Parameters
    ----------
    u : OccOrCT
        Multiplied oriented contig
    v : OccOrCT
        Multiplied oriented contig

    Returns
    -------
    bool
        True if (u, v) is an alpha key, False if (v, u) is an alpha key

    Raises
    ------
    KeyError
        If there is no key with u, v couple in alpha key set
    """
    if u[CIND_IND] < v[CIND_IND]:
        return True
    if u[CIND_IND] > v[CIND_IND]:
        return False
    # case u_id = v_id
    if u[COCC_IND] < v[COCC_IND]:
        return True
    if u[COCC_IND] > v[COCC_IND]:
        return False
    raise KeyError



# -*- coding=utf-8 -*-

"""Main when khloraascaff is launched in command line."""

from argparse import Namespace
from pathlib import Path

from khloraascaf.cli import (
    ARG_CONTIG_ATTRS,
    ARG_CONTIG_LINKS,
    ARG_CONTIG_STARTER,
    OPT_INSTANCE_NAME,
    OPT_MULT_UPB,
    OPT_OUTDEBUG,
    OPT_OUTDIR,
    OPT_PRESSCORE_UPB,
    OPT_SOLVER,
    arguments_cmdline,
)
from khloraascaf.scaffolding_methods import scaffolding


# ============================================================================ #
#                                     MAIN                                     #
# ============================================================================ #
if __name__ == '__main__':
    # ------------------------------------------------------------------------ #
    # Parse command line arguments
    # ------------------------------------------------------------------------ #
    args: Namespace = arguments_cmdline()

    # ------------------------------------------------------------------------ #
    # Run the scaffolding step
    # ------------------------------------------------------------------------ #
    output_dir_gen = scaffolding(
        Path(getattr(args, ARG_CONTIG_ATTRS)),
        Path(getattr(args, ARG_CONTIG_LINKS)),
        getattr(args, ARG_CONTIG_STARTER),
        multiplicity_upperbound=getattr(args, OPT_MULT_UPB),
        presence_score_upperbound=getattr(args, OPT_PRESSCORE_UPB),
        solver=getattr(args, OPT_SOLVER),
        outdir=Path(getattr(args, OPT_OUTDIR)),
        instance_name=getattr(args, OPT_INSTANCE_NAME),
        debug=getattr(args, OPT_OUTDEBUG),
    )
    #
    # To get the output directory path from shell
    #
    # TODO logging here output_dir_gen
    print(output_dir_gen)
